use std::ops::{Deref, DerefMut};

use jabber::{connection::Tls, jabber_stream::bound_stream::BoundJabberStream};
use read::ReadControlHandle;
use sqlx::SqlitePool;
use tokio::{
    sync::{mpsc, oneshot},
    task::{JoinHandle, JoinSet},
};
use write::{WriteControlHandle, WriteHandle, WriteMessage};

use crate::UpdateMessage;

mod read;
pub(crate) mod write;

pub struct Supervisor {
    connection_commands: mpsc::Receiver<SupervisorCommand>,
    writer_crash: oneshot::Receiver<(WriteMessage, mpsc::Receiver<WriteMessage>)>,
    reader_crash: oneshot::Receiver<(
        SqlitePool,
        mpsc::Sender<UpdateMessage>,
        tokio::task::JoinSet<()>,
    )>,
    sender: mpsc::Sender<UpdateMessage>,
    writer_handle: WriteControlHandle,
    reader_handle: ReadControlHandle,
    on_shutdown: oneshot::Sender<()>,
}

pub enum SupervisorCommand {
    Disconnect,
    // for if there was a stream error, require to reconnect
    Reconnect,
}

impl Supervisor {
    fn new(
        connection_commands: mpsc::Receiver<SupervisorCommand>,
        writer_crash: oneshot::Receiver<(WriteMessage, mpsc::Receiver<WriteMessage>)>,
        reader_crash: oneshot::Receiver<(SqlitePool, mpsc::Sender<UpdateMessage>, JoinSet<()>)>,
        sender: mpsc::Sender<UpdateMessage>,
        writer_handle: WriteControlHandle,
        reader_handle: ReadControlHandle,
        on_shutdown: oneshot::Sender<()>,
    ) -> Self {
        Self {
            connection_commands,
            writer_crash,
            sender,
            writer_handle,
            reader_handle,
            reader_crash,
            on_shutdown,
        }
    }

    async fn handle_command_message(&mut self, msg: SupervisorCommand) {}

    async fn run(mut self) {
        loop {
            tokio::select! {
                Some(msg) = self.connection_commands.recv() => {
                    self.handle_command_message(msg).await;
                },
                error = &mut self.writer_crash => {

                },
                error = &mut self.reader_crash => {

                },
                else => break,
            }
        }
        self.on_shutdown.send(());
    }
}

pub struct SupervisorHandle {
    sender: SupervisorSender,
    handle: JoinHandle<()>,
}

impl Deref for SupervisorHandle {
    type Target = SupervisorSender;

    fn deref(&self) -> &Self::Target {
        &self.sender
    }
}

impl DerefMut for SupervisorHandle {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.sender
    }
}

#[derive(Clone)]
pub struct SupervisorSender {
    sender: mpsc::Sender<SupervisorCommand>,
}

impl Deref for SupervisorSender {
    type Target = mpsc::Sender<SupervisorCommand>;

    fn deref(&self) -> &Self::Target {
        &self.sender
    }
}

impl DerefMut for SupervisorSender {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.sender
    }
}

impl SupervisorHandle {
    pub fn new(
        streams: BoundJabberStream<Tls>,
        update_sender: mpsc::Sender<UpdateMessage>,
        db: SqlitePool,
        on_shutdown: oneshot::Sender<()>,
    ) -> (WriteHandle, Self) {
        let (command_sender, command_receiver) = mpsc::channel(20);
        let (writer_error_sender, writer_error_receiver) = oneshot::channel();
        let (reader_crash_sender, reader_crash_receiver) = oneshot::channel();

        let (reader, writer) = streams.split();
        let (write_handle, write_control_handle) =
            WriteControlHandle::new(writer, writer_error_sender);
        let jabber_reader_control_handle = ReadControlHandle::new(
            reader,
            reader_crash_sender,
            db,
            update_sender.clone(),
            command_sender.clone(),
            write_handle.clone(),
        );

        let actor = Supervisor::new(
            command_receiver,
            writer_error_receiver,
            reader_crash_receiver,
            update_sender,
            write_control_handle,
            jabber_reader_control_handle,
            on_shutdown,
        );

        let handle = tokio::spawn(async move { actor.run().await });

        (
            write_handle,
            Self {
                sender: SupervisorSender {
                    sender: command_sender,
                },
                handle,
            },
        )
    }

    pub fn sender(&self) -> SupervisorSender {
        self.sender.clone()
    }
}
