use std::ops::{Deref, DerefMut};

use jabber::{connection::Tls, jabber_stream::bound_stream::BoundJabberWriter};
use stanza::client::Stanza;
use tokio::{
    sync::{mpsc, oneshot},
    task::JoinHandle,
};

use crate::error::Error;

// actor that receives jabber stanzas to write, and if there is an error, sends a message back to the supervisor then aborts, so the supervisor can spawn a new stream.
pub struct Write {
    stanza_receiver: mpsc::Receiver<WriteMessage>,
    control_receiver: mpsc::Receiver<WriteControl>,
    stream: BoundJabberWriter<Tls>,
    on_crash: oneshot::Sender<(WriteMessage, mpsc::Receiver<WriteMessage>)>,
}

pub struct WriteMessage {
    stanza: Stanza,
    respond_to: oneshot::Sender<Result<(), Error>>,
}

enum WriteControl {
    Disconnect,
    Abort(oneshot::Sender<mpsc::Receiver<WriteMessage>>),
}

impl Write {
    fn new(
        stanza_receiver: mpsc::Receiver<WriteMessage>,
        control_receiver: mpsc::Receiver<WriteControl>,
        stream: BoundJabberWriter<Tls>,
        supervisor: oneshot::Sender<(WriteMessage, mpsc::Receiver<WriteMessage>)>,
    ) -> Self {
        Self {
            stanza_receiver,
            control_receiver,
            stream,
            on_crash: supervisor,
        }
    }

    async fn write(&mut self, stanza: &Stanza) -> Result<(), peanuts::Error> {
        Ok(self.stream.write(stanza).await?)
    }

    async fn run(mut self) {
        loop {
            tokio::select! {
                Some(msg) = self.control_receiver.recv() => {
                    match msg {
                        WriteControl::Disconnect => {
                            // TODO: close the stanza_receiver channel and drain out all of the remaining stanzas to send
                            self.stanza_receiver.close();
                            // TODO: put this in some kind of function to avoid code duplication
                            for msg in self.stanza_receiver.recv().await {
                                let result = self.stream.write(&msg.stanza).await;
                                match result {
                                    Err(e) => match &e {
                                        peanuts::Error::ReadError(error) => {
                                            // make sure message is not lost from error, supervisor handles retry and reporting
                                            self.on_crash.send((msg, self.stanza_receiver));
                                            break;
                                        }
                                        _ => {
                                            msg.respond_to.send(Err(e.into()));
                                        }
                                    },
                                    _ => {
                                        msg.respond_to.send(Ok(()));
                                    }
                                }
                            }
                            self.stream.try_close().await;
                            break;
                        },
                        WriteControl::Abort(sender) => {
                            sender.send(self.stanza_receiver);
                            break;
                        },
                    }
                },
                Some(msg) = self.stanza_receiver.recv() => {
                    let result = self.stream.write(&msg.stanza).await;
                    match result {
                        Err(e) => match &e {
                            peanuts::Error::ReadError(error) => {
                                // make sure message is not lost from error, supervisor handles retry and reporting
                                self.on_crash.send((msg, self.stanza_receiver));
                                break;
                            }
                            _ => {
                                msg.respond_to.send(Err(e.into()));
                            }
                        },
                        _ => {
                            msg.respond_to.send(Ok(()));
                        }
                    }
                },
                // TODO: check if this is ok to do
                else => break,
            }
        }
    }
}

#[derive(Clone)]
pub struct WriteHandle {
    sender: mpsc::Sender<WriteMessage>,
}

impl Deref for WriteHandle {
    type Target = mpsc::Sender<WriteMessage>;

    fn deref(&self) -> &Self::Target {
        &self.sender
    }
}

impl DerefMut for WriteHandle {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.sender
    }
}

pub struct WriteControlHandle {
    sender: mpsc::Sender<WriteControl>,
    handle: JoinHandle<()>,
}

impl WriteControlHandle {
    pub fn new(
        stream: BoundJabberWriter<Tls>,
        supervisor: oneshot::Sender<(WriteMessage, mpsc::Receiver<WriteMessage>)>,
    ) -> (WriteHandle, Self) {
        let (control_sender, control_receiver) = mpsc::channel(20);
        let (stanza_sender, stanza_receiver) = mpsc::channel(20);

        let actor = Write::new(stanza_receiver, control_receiver, stream, supervisor);
        let handle = tokio::spawn(async move { actor.run().await });

        (
            WriteHandle {
                sender: stanza_sender,
            },
            Self {
                sender: control_sender,
                handle,
            },
        )
    }

    pub fn reconnect(
        stream: BoundJabberWriter<Tls>,
        supervisor: oneshot::Sender<(WriteMessage, mpsc::Receiver<WriteMessage>)>,
        stanza_receiver: mpsc::Receiver<WriteMessage>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = Write::new(stanza_receiver, control_receiver, stream, supervisor);
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }
}
